 -------------------------- MODULE Actions ---------------------------
\* This module defines all the actions

(* The hierarchy of the modules are
 Ring (The basic parameters and operations for ring calculation)
   \subset LS (The leaf set utilities)
      \subset RT (The routing table utilities)
         \subset Msg (The messages definitions)
            \subset InitialStates (The initial state and TypeInvariant)
               \subset Actions (All the actions)
                  \subset MSPastry (The specification of MSPastry including propertie
                  \subset MCMSPastry (The model checking file) 
*)
EXTENDS InitialStates
---------------------------------------------------------------

\* Application relevant actions

\* an application on node i tries to find the node j 
\* or to find the node which is closest to the key j.
\* The node i generate a Lookup msg and ask himself to
\* route it to j. 
Lookup(i, j) ==
\*  /\ ~\E m \in receivedMsgs: 
\*     /\ m.mreq.type = "Lookup"
\*     /\ m.mreq.node = j
  /\ receivedMsgs'= receivedMsgs \cup 
         {[destination |-> i, 
                  mreq |-> [type |-> "Lookup", 
                            node |-> j]]}
  /\ UNCHANGED <<status, lset, probing, 
                      failed, lease, grant, toj>>


\* a node i has got an ID "seed" from the application running based
\* on the overlay network and he wants to join the network.
\* Seed is one of the status nodes in the network who is in charge of
\* some range of keys which is numerrically close to him.
\* The node i generate the JoinRequest msg for seed to let him
\* route the msg to the ID i. By doing this, the msg will be replied
\* by the node nummerically closest to i and hence find the correct
\* place for i to join into the network.
Join(j, i) == 
\*  /\ i # seed \* can be derived from status[i] = "dead" and status[seed] ="ready"        
\*  /\ ~\E m \in receivedMsgs: 
\*     /\ m.mreq.type = "JoinRequest"
\*     /\ m.mreq.node = i    
  /\ status[j] = "dead"
  /\ status[i] ="ready"        
\*  /\ lset[i] = EmptyLS(i)    can be derived.       
  /\ status' = [status EXCEPT ![j]="wait"]
  /\ receivedMsgs'= receivedMsgs \cup 
        {[destination |-> i, 
          mreq |-> [type |-> "JoinRequest", 
                       node |-> j]
             ]}
  /\ UNCHANGED <<lset, probing, failed, 
                  lease, grant, toj>>
------------------------------------------------
\* Internal behavior of system 

\* (1) Delivering 
\* a node i deliver the lookup message it take charge of k
\* Only the status node can deliver Lookup messages,
\* which destination is leaseed by this node
Deliver(i, j) == 
  /\ status[i] = "ready" 
  /\ \E m \in receivedMsgs: 
     /\ m.mreq.type = "Lookup"
     /\ m.destination = i
     /\ m.mreq.node = j
     /\ Covers(lset[i], j)
     /\ receivedMsgs'= (receivedMsgs \{m}) 
  /\ UNCHANGED <<status, lset, probing, failed, lease, grant, toj>>
---------------------------------------------------------
\* (2) Routing

\*Help function for route
FindNext(i, j) == 
\* very simple version
  LET 
     lsCan ==  GetLSetContent(lset[i]) \ failed[i]      
  IN IF lsCan # {i}
     THEN LET next == CHOOSE n \in (lsCan \ {i}): 
                      \A m \in lsCan: AbsDist(n, j) =< AbsDist(m, j)
          \* to handle the case when there is another node n in leafset
          \* which has the same abs distance with j as i do.
          \* since FindNext is only used when the node i does not covers 
          \* the node, the other node n is always chosen to forward the message
          IN IF AbsDist(next, j) = AbsDist(i, j) THEN next
             ELSE CHOOSE n \in lsCan: 
                      \A m \in lsCan: AbsDist(n, j) =< AbsDist(m, j)
     ELSE i
(* simplified version
  LET 
     lsCan ==  GetLSetContent(lset[i]) \ failed[i]
     canrelax ==
           {can \in (GetLSetContent(lset[i]) 
                    \cup GetRTableContent(rtable[i]))
                    \ failed[i]:
                       AbsDist(j, can) < AbsDist(j, i)}
  IN 
   
    IF \* We should find the next hop in leaf set
     \* when the leaf set of i covers whole ring,
     \* or i is within the range of leaf set;
      \/ Overlaps(lset[i])      
      \/ CwDist(LeftMost(lset[i]), j) =< CwDist(LeftMost(lset[i]), RightMost(lset[i]))
     \*/\ lsCan # {} \* this set cannot be empty because it contains i which cannot be 
                   \* inside failed[i]
    THEN IF lsCan # {i}
         THEN LET next == CHOOSE n \in (lsCan \ {i}): 
                      \A m \in lsCan: AbsDist(n, j) =< AbsDist(m, j)
          \* to handle the case when there is another node n in leafset
          \* which has the same abs distance with j as i do.
          \* since FindNext is only used when the node i does not covers 
          \* the node, the other node n is always chosen to forward the message
              IN IF AbsDist(next, j) = AbsDist(i, j) 
                 THEN next
                 ELSE CHOOSE n \in lsCan: 
                      \A m \in lsCan: AbsDist(n, j) =< AbsDist(m, j)
         ELSE i
  \* Real network will check the liveness via underlying TCP PING, but here
  \* we assume the TCP is reliable and we pretend to get the liveness
    ELSE \*Otherwise we should try the rtable
       IF canrelax # {} 
       THEN CHOOSE can \in canrelax: \A m \in canrelax: AbsDist(can, j) =< AbsDist(m, j)
       ELSE i

       
       
       
\* The following is the complex algorithm involves routing table lookup
(* To get the next forward destination.*)
\* since we check the liveness of a node before pick it as candidate, 
\* we don't take it if it is "dead"
  IF \* We should find the next hop in leaf set
     \* when the leaf set of i covers whole ring,
     \* or i is within the range of leaf set;
     \/ Overlaps(lset[i])      
     \/ CwDist(LeftMost(lset[i]), j) =< CwDist(LeftMost(lset[i]), RightMost(lset[i]))
  THEN LET lsCan == GetLSetContent(lset[i]) \ failed[i] \* if leafset of i is empty then i \in lsCan
   \*{c \in GetLSetContent(lset[i]) \ failed[i]: status[c] # "dead"}
  \* Real network will check the liveness via underlying TCP PING, but here
  \* we assume the TCP is reliable and we pretend to get the liveness
       IN IF lsCan # {i}
          THEN LET next == CHOOSE n \in (lsCan \ {i}): 
                      \A m \in lsCan: AbsDist(n, j) =< AbsDist(m, j)
          \* to handle the case when there is another node n in leafset
          \* which has the same abs distance with j as i do.
          \* since FindNext is only used when the node i does not covers 
          \* the node, the other node n is always chosen to forward the message
              IN IF AbsDist(next, j) = AbsDist(i, j) 
                 THEN next
                 ELSE CHOOSE n \in lsCan: 
                      \A m \in lsCan: AbsDist(n, j) =< AbsDist(m, j)
          ELSE i
  ELSE \*Otherwise we should try the rtable
       LET r == SharedDig(j, i)
           rtCan == rtable[i][<<r, GetDig(j, r+1)>>]
       IN IF rtCan # null \*/\ status[rtCan] # "dead"
          THEN rtCan
          ELSE \* If we cannot find anyone from rtable, we relax the condition and try with 
               \* all we know from rtable and lset
               LET canrelax ==
                   {can \in (GetLSetContent(lset[i]) 
                            \cup GetRTableContent(rtable[i]))
                            \ failed[i]:
                   AbsDist(j, can) < AbsDist(j, i) /\ SharedDig(j, can) >= r}\* /\ status[can] # "dead"}
                   \* again, we simulate the check liveness here
               IN IF canrelax # {} 
                  THEN CHOOSE can \in canrelax: \A m \in canrelax: AbsDist(can, j) =< AbsDist(m, j)
                  ELSE i
       *)

\* the node i forwards a look up to a closer node to the final destination k
\* when the destination key is outside of its range
\* If we can't find the closer node, then we fire an Error
RouteLookup(i, j) == 
  \*/\ status[i] = "ready"
  /\ status[i] # "dead"
  /\ \E m \in receivedMsgs:
     /\ m.mreq.type ="Lookup" 
     /\ m.destination = i
     /\ m.mreq.node = j 
     /\ ~Covers(lset[i], j)
     /\ LET nh ==  FindNext(i, j)
        IN  receivedMsgs'= (receivedMsgs \{m}) \cup 
          IF nh # i
          THEN {[destination |-> nh, mreq |-> m.mreq]}
          ELSE {[destination |-> i, mreq |-> 
                [type |-> "NoLegalRoute", key |-> j]]}
  /\ UNCHANGED <<status, lset, probing, 
                 failed, lease, grant, toj>>
          
\* the node i forwards a join request to a closer node to the final destination k
\* when the destination key is outside of its range
\* If we can't find the closer node, then we fire an Error
RouteJReq(i, j) == 
  \*/\ status[i] = "ready"
  /\ status[i] # "dead"
  /\ \E m \in receivedMsgs:
     /\ m.mreq.type = "JoinRequest"
     /\ m.destination = i
     /\ m.mreq.node = j 
     \* /\ ~k \in GetLSetContent(lset[i]) 
        \* to avoid the case that a node left and rejoin 
        \* before its neighbor deleted it from their leaf set
        \* but this should not cause a real problem. If I block the join, then the rejoined node will never be able to join.
     /\ ~Covers(lset[i], j)
     /\ LET nh ==  FindNext(i, j)
        IN 
        receivedMsgs'= (receivedMsgs \{m}) \cup 
          IF nh # i
          THEN {[destination |-> nh, mreq |-> 
                [type |-> "JoinRequest",
                 node |-> j]]}
               \*for join request, we update the rtable by each hop
          ELSE {[destination |-> i, mreq |-> 
                [type |-> "NoLegalRoute", key |-> j]]}
  /\ UNCHANGED <<status, lset, probing, failed, lease, grant, toj>>
-----------------------------------------------------------------------------------------------
\* (3) Consistent Join

\* node i receives join request msg and send back join reply with its leaf set
\* when the node i is "ready" and the destination is within the cover range
RecJReq(i) ==  
  /\ status[i] = "ready" 
  /\ toj[i] = i \* to ensure that a ready node can handle maximal one node to join at a time
  /\ \E m \in receivedMsgs:
   /\ m.mreq.type = "JoinRequest"
   \*/\ m.mreq.node # i can be derived 
   /\ m.destination = i
   \* /\ ~m.mreq.node \in GetLSetContent(lset[i]) 
   \* to avoid the case that a node left and rejoin 
   \* before its neighbor deleted it from their leaf set
   \* but this should not cause a real problem. If I block the join, then the rejoined node will never be able to join.
   /\ Covers(lset[i], m.mreq.node)
   /\ LET newmreq == [type |-> "JoinReply",  
                              lset |-> lset[i]]
          newm == [destination |-> m.mreq.node,
                           mreq |-> newmreq]
      IN receivedMsgs'= (receivedMsgs \{m}) \cup {newm}
   /\ toj' = [toj EXCEPT ![i]= m.mreq.node]
   /\ lset' = [lset EXCEPT ![i] = AddToLSet({m.mreq.node}, lset[i])]
  /\ UNCHANGED <<status, probing, failed, 
                 lease, grant>>
   
\* A function to build the Probe Message to its neighbour
ProbeSet(i, ls, toprob) == 
  {[destination |-> j,
          mreq |-> [type |-> "Probe",  
                    node |-> i,
                    lset |-> ls
                   ] ]: j \in toprob}
  
\* Node i receives join reply msg, it updates its rtable and lset with
\* the received information and probe the new nodes in its leaf set.
RecJReply(i) == 
/\ status[i] = "wait" 
\*/\ probing[i] = {}
\*/\ lset[i] = EmptyLS(i) \* we assume that when a node leaves the network, 
                          \* all the messages towards it will be deleted.
                          \* Hence, there is at most one join reply to any node at
                          \* any time (Invariant: OneJRply). Hence, these conditions
                          \* can be derived from the followed condition.
/\ \E m \in receivedMsgs: 
   /\ m.mreq.type = "JoinReply"
   /\ m.destination = i
\*   /\ m.mreq.lset.node # i \* based on the fact that a node will never send a join reply
                             \* to itself, this could be derived.
   /\ LET

        newlset == AddToLSet(GetLSetContent(m.mreq.lset), lset[i])
        toprob  == GetLSetContent(newlset) \ {i}\*, m.mreq.lset.node}
        \* without probing it cannot become OK
        probmsgs== ProbeSet(i, newlset, toprob)
      IN
     /\ lset'   = [lset    EXCEPT ![i] = newlset] 
     /\ probing'= [probing EXCEPT ![i] = toprob]
     \*/\ failed' = [failed  EXCEPT ![i] = {}] \* should be anyway empty.
     /\ receivedMsgs'= (receivedMsgs \ {m}) \cup probmsgs
/\ UNCHANGED <<status, lease, grant, toj, failed>>

RecProbeInner(i, m) ==
LET j == m.mreq.node
        fi == failed[i] \ {j} 
       ls1 == AddToLSet({j}, lset[i]) 
       \*ls2 == RemoveFromLSet(m.mreq.failed, ls1)
       \* we don't remove any node from leaf set before node i's lease finishes.
     lprim == GetLSetContent(AddToLSet((GetLSetContent(m.mreq.lset) \ fi), ls1))\* see above
      
      prb2 == lprim \ GetLSetContent(ls1) \* need to confirm the new candidates for the leaf set are real alive
      \* before we want to add them
      prb == prb2 \ (probing[i] \cup fi) 
      newm == [type |-> "ProbeReply",  
                  node |-> i,
                  lset |-> lset[i]]
     IN
         /\failed' = [failed EXCEPT ![i]=fi]
         /\lset'   = [lset EXCEPT ![i] = ls1]
         /\probing'= [probing EXCEPT ![i] = @ \cup prb] 
         \* If node i has probed j and waiting its reply but then get the probe msg from j
         \* it continue to wait j's reply. So, the probing' might contains j 
         
         \* we delete the change of status status because we don't allow the remove of leaf set member here.
         \*/\status' = [status EXCEPT ![i] = IF ~IsComplete(ls2) /\ @ \in {"ready", "ok"} THEN "wait" ELSE @]
         \* In case we removed the nodes of m.mreq.failed from leaf set of i, then it is no more complete
         \* we need to change the node's status in order to complete the leaf set first before it can route badly
         \* see vE-pastry/log/log-RecProbeWithFailedNodes.txt since step 19, when node 17 received msg from 18
         \* containing the failed node 95, node 17 removes 95 from its leaf set, then it is no more completed
         /\receivedMsgs' = (receivedMsgs  
                           \cup { [destination |-> j, mreq |-> newm]} 
                           \cup ProbeSet(i, ls1, prb)
                           ) \ {m}
/\ UNCHANGED <<lease, status, grant, toj>>
\* Node i receives probe msg from j. 
\* It updates the local knowledge w.r.t. j and send reply back to j.
\* It probes nodes in leaf set which are in failed and removes them from its leaf set.(prb1)
\* Then it creates the clone L' of its leaf set and add nodes in L, which it 
\* consideres not to be faulty. (lprim)
\* It probes the new nodes in lprim before inclusion. (prb2)
RecProbe(i) ==
/\ status[i] # "dead"
/\ status[i] = "ready" \/ lset[i] # EmptyLS(i)
\*a \wait node $j$, which has just sent its \mjrq, 
\* will not react on any \mprb messages that  come from a neighboring node $n$ of the potential replier $i$
\* to its \mjrq message. Due to message delay, the probing message from node $n$ could arrive the node $j$ earlier than its 
\* waiting \mjrp message from the node $i$. This might disturb the normal join process and cause malicious behavior and inconsistency. 
/\ \E m \in receivedMsgs:
  /\ m.mreq.type = "Probe"
  /\ m.destination = i
  \*/\ m.mreq.node = m.mreq.lset.node \* can be derived from SemPRply
  \*/\ m.mreq.node # i \* can be derived
  /\ RecProbeInner(i, m) 

\* A node i receives probe reply msg. It does similar things as receiving probe msg except:
\* instead of sending reply, it removes j from probing; (prb3)
\* If there is still out standing probes, do nothing. (prb3 = {})
\* If no out standing probes exists, it checks if its leaf set is complete
\* when yes, it turns its status to "ok" and thus finish the consistent join phrase.
\* If the leaf set is not complete, it probes the farest nodes (leftmost and rightmost) 
\* in leaf set to complete it.
\* lprim : potential new leaf set by adding received leaf set.
\* prb1:  heard of some failed nodes, which are my leaf set member, 
\*        need to be confirmed before remove
\* prb2:  brand new nodes in the potential new leaf set need first 
\*        to be probed then be added.
\* prb3:  overall all probing nodes after this action
\* A node has no outstanding probe and has completed its leaf set
\* then it should turn to be status ("ok")
RecPRplyInner(i,m) ==
  LET    j == m.mreq.node
        Ls == m.mreq.lset
        fi == failed[i] \ {j} 
       ls1 == AddToLSet({j}, lset[i]) 
     lprim == GetLSetContent(AddToLSet((GetLSetContent(Ls) \ fi), ls1))
      prb2 == lprim \ (GetLSetContent(ls1) \cup probing[i] \cup fi)
      prb3 == (probing[i] \cup prb2) \ {i, j} 
shouldBeOK == /\ status[i] = "wait"
              \*/\ PrintT(<<i, status[i], prb3>>) 
              /\ prb3={}
              \*/\ Overlaps(ls1)\/IsComplete(ls1)\* this is only useful when we allow
                                               \* node departure, when the node have 
                                               \*removed another node from one side of its leaf set
                                               \* in this case, it is neither complete nor overlaping.
                                               \* But in that case, node i must have started 
                                               \* leafset repair by probing other nodes.
  IN 
     /\ lset'   = [lset    EXCEPT ![i] = ls1]
     /\ failed' = [failed  EXCEPT ![i] = IF shouldBeOK THEN {} ELSE fi]       
     /\ probing'= [probing EXCEPT ![i] = prb3]
     /\ status' = [status  EXCEPT ![i] = IF shouldBeOK THEN "ok" ELSE @]      
     /\ receivedMsgs' = (receivedMsgs 
                        \cup ProbeSet(i, ls1, prb2)
                        \cup IF shouldBeOK
                             THEN {[destination |-> LeftNeighbor(ls1), 
                                     mreq |-> [type |-> "LeaseRequest",
                                               node |-> i]],
                                   [destination |-> RightNeighbor(ls1),
                                     mreq |-> [type |-> "LeaseRequest",
                                               node |-> i]]
                                               }
                             ELSE {}) 
                        \ {m} 
     /\ UNCHANGED <<lease, grant, toj>>

\* only "dead" nodes donnot react on probe reply.
\* others' faked probe-reply might make i status
\* and hence create a partition of network
RecPRply(i) ==
/\ status[i] # "dead"
\*/\ lset[i] # EmptyLS(i) \* can be derived
/\ \E m \in receivedMsgs:
   /\ m.mreq.type = "ProbeReply"  
   /\ m.destination = i
   \*/\ m.mreq.node \in probing[i] \* can be derived
   \*/\ m.mreq.node # i \* can be derived
   /\ RecPRplyInner(i,m)
  
\* A node i suspected j to be faulty and hence probe j
SuspectFaulty(i, j) == 
  /\ status[i] \in {"ok", "ready"} \* it doesn't make sense for a waiting node to suspect any dead node
       \* By enhencing this condition, we are able to say, i cannot be
       \* the single ready/ok node on the ring, since its leaf set is not empty.
  /\ status[j]= "dead" \* For simplicity we enforce it. 
  /\ i # j \* node doesnot suspect itself
  /\ j \in GetLSetContent(lset[i]) \* node only suspect nodes in its leaf set
  /\ j \notin probing[i] \*node does not suspect outstanding probed node
  /\ j \notin failed[i] \*node does not suspect a 'dead' node it knows
  /\ receivedMsgs'= receivedMsgs \cup ProbeSet(i, lset[i], {j})
  /\ probing'=[probing EXCEPT ![i]= @\cup {j}]
  /\ UNCHANGED <<lset, failed, lease, grant, status, toj>>


ChooseCandidates(ls)
==LET 
    lm == {LeftMost(ls)} \ {ls.node}
    rm == {RightMost(ls)} \ {ls.node}
  IN 
    IF Lenth(ls.left) < L /\ Lenth(ls.right) < L
    THEN lm \cup rm
    ELSE IF Lenth(ls.left) < L /\ Lenth(ls.right) = L
         THEN IF lm # {} THEN lm ELSE rm
         ELSE IF rm # {} THEN rm ELSE lm
\* Simulation of a normal behavior.
\* A node i stops waiting reply from node j and considers j to be failed  
\* In the real system, 
\* the node i will try several times before it gives up waiting. 
\* Besides it will try different ways
\* to contact node j, including using virtual link techniques.
\* We simplified the virtual link technique in our model:
\* If a node has blocked incoming channel but not outgoing channel,
\* Then we assume that a virtual link will be set up so that it behaves
\* exactly like a normal node. Hence, after using virtual link, 
\* the node can contact each other. Hence all the probe will be ordinarily handled.
\* In all, as soon as probe time out happend by i for probing j, it means
\* there is no way for i to find j, even through virtual link. 
\* Then there is only two possibilities: either i has already left the
\* network or j has. 
\* The first case is ruled out by previous gard, so j must have left the network.
\* That's why we simulate the real situation like this. 
TimeOut(i, j)== 
/\ status[i]# "dead" \* for nodes no more in the network, it does not matter
/\ \/ status[j]= "dead" 
   \/ /\ status[j]= "wait"
      /\ lset[j] = EmptyLS(j)\* For simplicity we enforce it. 
/\ j \in probing[i]
/\ failed' = [failed EXCEPT ![i] = @\cup {j}]
/\ lset' = [lset EXCEPT ![i] = IF ~(j \in grant[i]) /\ j \in GetLSetContent(@)
                               THEN RemoveFromLSet({j},@) 
                               ELSE @]
   \* A fix after log/vI-Pastry/HN-NC-newProbe-newGrant.log, to prevent a node becoming 
   \* wait when some previous leafset node left the network. Now, only if someone from 
   \* the current leaf set left the network, this node will change its status
/\ probing' = [probing EXCEPT ![i] = (@ \ {j}) \cup 
                               IF ~(j \in grant[i]) /\ j \in GetLSetContent(lset[i])
                               THEN ChooseCandidates(RemoveFromLSet({j},lset[i]))
                               ELSE {}]     
/\ receivedMsgs' = (receivedMsgs \
                          IF \E m \in receivedMsgs: m.mreq.type = "Probe"
\* here we delete all the gabage probe msgs from i to j, since 
\* we are sure no one will get a reply any more. 
                                /\ m.mreq.node = i
                                /\ m.destination = j
                          THEN  {m \in receivedMsgs: m.mreq.type = "Probe"
                                   /\ m.mreq.node = i
                                   /\ m.destination = j}
                          ELSE {})
                   \cup IF ~(j \in grant[i]) /\ j \in GetLSetContent(lset[i])
                        THEN ProbeSet(i, lset[i], ChooseCandidates(RemoveFromLSet({j},lset[i])))
                        ELSE {}
/\ status'= [status EXCEPT ![i] = IF ~(j \in grant[i]) /\ j \in GetLSetContent(lset[i]) THEN "wait" ELSE @] 
/\ lease'= [lease EXCEPT ![i] = @ \ {j}]
/\ UNCHANGED <<grant, toj>> 
-----------------------------------------------------------------------------------------------------
\* (4) lease protocol

RequestLease(i) ==
    \*/\ status[i] \in {"ok", "ready"} 
    /\ status[i] ="ok"
    /\ LET ln == LeftNeighbor(lset[i])
           rn == RightNeighbor(lset[i])
       IN /\ ~ (ln \in lease[i] /\ rn \in lease[i])
          /\ receivedMsgs' = (receivedMsgs 
             \cup IF ~ (ln \in lease[i])
                  THEN {[destination |-> ln, 
                         mreq |-> [type |-> "LeaseRequest",
                                   node |-> i]]}
                  ELSE {})
             \cup IF ~ (rn \in lease[i])
                  THEN {[destination |-> rn, 
                         mreq |-> [type |-> "LeaseRequest",
                                   node |-> i]]}
                  ELSE {}
    /\ UNCHANGED <<status, lset, probing, 
                     failed, lease, grant, toj>>
\* A node received a LeaseRequest msg, 
\* if it is from neighbor, then it grands the lease and send reply back
\* if not from its neighbor, it deletes the request msg   
RecLReq(i) ==  
\*/\ status[i] = "ready"
/\ status[i] \in {"ok", "ready"}   
\*/\ lset[i] # EmptyLS(i) \* can be derived
/\ \E m \in receivedMsgs:
   /\ m.mreq.type = "LeaseRequest" 
   /\ m.destination = i
   \*/\ m.mreq.node # i\* can be derived from SemLReq
   \* There are two cases the request is taken care of: 1. a new node join
   \* 2. the lease has expired and it request it for a new lease
   \* For both cases, we grand a lease
   /\ grant' =[grant EXCEPT ![i] = 
                IF m.mreq.node \in {LeftNeighbor(lset[i]), 
                                    RightNeighbor(lset[i])}
                THEN @ \cup {m.mreq.node} 
                ELSE @]
   /\ receivedMsgs' = (receivedMsgs \ {m})
       \cup {[destination |-> m.mreq.node, 
               mreq |-> [type |-> "LeaseReply",
                         lset |-> lset[i],
                         grant|-> 
                    m.mreq.node \in {LeftNeighbor(lset[i]), 
                                         RightNeighbor(lset[i])}]]}
/\ UNCHANGED <<status, lset, probing, failed, lease, toj>>      

\* A node receives LeaseReply msg. from its neighbors 
\* if it get the grant, then it add the lease of the neighbor
\* if the node got leases from both of its neighbors, it turns to be Ready
RecLReply(i) ==  
/\ status[i] \in {"ready", "ok"}  
\*/\ lset[i] # EmptyLS(i)
/\ \E m \in receivedMsgs:
   /\ m.mreq.type = "LeaseReply" 
   /\ m.destination = i   
   /\ m.mreq.lset.node \in {LeftNeighbor(lset[i]), 
                            RightNeighbor(lset[i])}
   /\ LET 
         ln == LeftNeighbor(lset[i])
         rn == RightNeighbor(lset[i])
         newlease == [lease EXCEPT ![i] = 
                       IF m.mreq.grant = TRUE
                       THEN @ \cup {m.mreq.lset.node}
                       ELSE @]
         okToReady == /\ ln \in newlease[i] 
                      /\ rn \in newlease[i]
                      /\ status[i] = "ok"
      IN 
      /\ lease' = newlease
      /\ toj'=[toj EXCEPT ![i]= IF toj[i]=m.mreq.lset.node 
                                THEN i
                                ELSE @]
      /\ status' = [status EXCEPT ![i] =   IF okToReady
                                           THEN "ready"
                                           ELSE @ ]
      /\ grant' = [grant EXCEPT ![i] 
                   = IF okToReady 
                     THEN @ \cup {LeftNeighbor(lset[i]), RightNeighbor(lset[i])}
                     ELSE @]                                           
      /\ receivedMsgs' = IF okToReady 
                         THEN (receivedMsgs \ {m})
                       \cup {[destination |-> LeftNeighbor(lset[i]), 
                                       mreq |-> [type |-> "LeaseReply",
                                                 lset |-> lset[i],
                                                 grant|-> TRUE]]} 
                                                 \* So that the ready neighbor release the toj
                               \cup {[destination |-> RightNeighbor(lset[i]), 
                                       mreq |-> [type |-> "LeaseReply",
                                                 lset |-> lset[i],
                                                 grant|-> TRUE]]}                              
                         ELSE (receivedMsgs \ {m})                         
/\ UNCHANGED <<lset, probing, failed>>
       
\* One of a node's lease has expired after a fixed time period T after it has requested for a lease.
\* A node starts to count the time down for lease expiration in two different cases: 
\* (1). it initially requested its neighbor for a lease
\* (2). it received the request and then grand the lease for its neighbor.
\* Since there is an order of this two cases, (1) must happen before (2) for a pair of neighbors,
\* the lease expires also in the same order. 
\* Therefore, whenever the node's lease is in the situation of 2, we need to make sure its 
\* neighbor has already expired the lease in our simulation.
\* the lease of node j has expired by i.  This is done with the following two actions.
LeaseExp(i, j) == 
  \*/\ \E k \in I: k#i /\ status[k]="ready" \*if this is the only ready node, its lease cannot expire
  \* even if there is only one node ready on the network, it should expire its dead neighbor's lease
  /\ status[i] \in {"ready", "ok"}
  (*
  /\ \/ lease[i][j] = 1 \*after node i has got grand from its neighbor j for a time period
     \/ /\ lease[i][j] = 2 \*or after node i has granded lease to its neighbor j for a time period
        /\ lease[j][i] = 0     \*and the node j's lease of i has already expired
  /\ lease' = [lease EXCEPT ![i][j] = 0]
  *) 
  /\ i # j
  /\ j \in lease[i]
  /\ lease' = [lease EXCEPT ![i] = @ \{j}]
  /\ status' = [status EXCEPT ![i] = IF j \in {LeftNeighbor(lset[i]), RightNeighbor(lset[i])} /\ @ = "ready"
                                     THEN "ok" 
                                     ELSE @]
  /\ receivedMsgs' = receivedMsgs \cup
                        {[destination |-> j, 
                                     mreq |-> [type |-> "LeaseRequest",
                                               node |-> i]]}                                 
  /\ UNCHANGED <<lset, probing, failed, grant, toj>>
  
\* This is a simulation of the order of the lease expiration. (see above action's comments)
GrantExp(i, j) == 
/\ i # j
/\ status[i] \in {"ready", "ok"}
/\ j \in grant[i] /\ ~ (i \in lease[j])
/\ grant'= [grant EXCEPT ![i] = @ \ {j}]
/\ lset' = [lset EXCEPT ![i] = IF j \in failed[i] THEN RemoveFromLSet({j},@) ELSE @]
/\ status'= [status EXCEPT ![i] = IF j \in failed[i] THEN "wait" ELSE @]
/\ probing' = [probing EXCEPT ![i] = 
                   IF j \in failed[i] 
                   THEN probing[i] \cup ChooseCandidates(RemoveFromLSet({j},lset[i])) 
                   ELSE @]
/\ receivedMsgs' = receivedMsgs \cup 
                   IF j \in failed[i] 
                   THEN ProbeSet(i, lset[i], ChooseCandidates(RemoveFromLSet({j},lset[i])))     
                   ELSE {}               
/\ toj'= [toj EXCEPT ![i] = IF @ = j THEN i ELSE @]
/\  UNCHANGED <<lease, failed>>                  


\* node i finally removes its critical neighbor j from its leaf set. 
\* if there is no more possible repair for the node, it will
\* try to resign to the network
(*
DeclareDead(i,j) == 
/\ status[i] \in {"ok", "ready"}
/\ j \in failed[i] \* After i has put j in failed for a while.
/\ j \in GetLSetContent(lset[i])   \* and it is still in leaf set
/\ j = LeftNeighbor(lset[i]) => ~ j \in grant[i]
/\ j = RightNeighbor(lset[i])=> ~ j \in grant[i]
\* If j is i's neighbor, then its lease must be already expired before 
\* it can be declared dead.
/\ lset' = [lset EXCEPT ![i] = RemoveFromLSet({j},@)]
/\ status'= [status EXCEPT ![i] = "wait"] 
/\ probing' = [probing EXCEPT ![i] =  probing[i] \cup ChooseCandidates(RemoveFromLSet({j},lset[i]))]
/\ receivedMsgs' = receivedMsgs \cup ProbeSet(i, lset[i], failed[i], ChooseCandidates(RemoveFromLSet({j},lset[i])))
/\ UNCHANGED <<lease, failed, grant, toj>>
*)
(*
ResignNode(i) == \* need to resign and rejoin
\* when half of the leaf set nodes drop
/\ status[i] = "wait"
/\ probing[i] = {}
/\ Lenth(lset[i].left) = 0  \/ Lenth(lset[i].right) = 0
/\ ~\E m \in receivedMsgs: m.mreq.type = "JoinReply" /\ m.destination = i 
/\ \E n \in I: n#i /\ status[n] = "ready"
    /\ lset'= [lset EXCEPT ![i] = EmptyLS(i)]
    /\ probing' = [probing EXCEPT ![i] =  {}]
    /\ lease' = [lease EXCEPT ![i] = {}]
    /\ grant' =  [grant EXCEPT ![i] =  {}]
    /\ failed' = [failed EXCEPT ![i] = {}]
    /\ receivedMsgs' = receivedMsgs \cup 
                      {[destination |-> n, 
                        mreq |-> [
                            type |-> "JoinRequest", 
                           node |-> i]
                       ]}
/\ UNCHANGED <<status, toj>>

Recover(i)== \* when resign is needed but no one could be found to resign,
\* i will recover the network by probing the farest element in its leaf set.
/\ status[i] = "wait"
/\ probing[i] = {}
/\ Lenth(lset[i].left) = 0  \/ Lenth(lset[i].right) = 0 
/\ ~\E n \in I:  status[n] \in {"ready", "ok"}
/\ ~(Lenth(lset[i].left) = 0  /\ Lenth(lset[i].right) = 0) 
\* The "ok" nodes are the first candidate to set up the network than this one
\* so as long as there is "ok" node, this node should wait till the "ok" node turns
\* "ready" see log-concurrent-leave-last.txt in vE...
/\ LET 
      pe == IF Lenth(lset[i].left) = 0 
            THEN RightMost(lset[i])
            ELSE LeftMost(lset[i])
   IN 
   /\ probing' = [probing EXCEPT ![i] =  probing[i] \cup {pe}]
   /\ receivedMsgs' = receivedMsgs \cup ProbeSet(i, lset[i],  {pe})
/\ UNCHANGED <<lset,failed, lease, status, grant, toj>>


\* if all the nodes go to wait or dead, one must be able to be ready again.
\* to ensure the assumption that at least one node is ready on the network.
\* This action is a simulation to a server restart, only the oracle can do this.
\* Because other waiting nodes, who has more information could trigger the 
\* recover action. So as long as there is some node being able to recover,
\* we do not restart.
Restart(i)== \* when resign is needed but no one could be found to resign,
\* and no one can be found to recover
\* i will restart the network and select itself to be the only ready one
/\ status[i] = "wait"
/\ probing[i] = {}
\* The following sentence contains the information only from the oracle
/\ \A k \in NonDead: Lenth(lset[k].left) = 0  /\ Lenth(lset[k].right) = 0 
/\ ~\E n \in I:  status[n] \in {"ready", "ok"}
\* The "ok" nodes are the first candidate to set up the network than this one
\* so as long as there is "ok" node, this node should wait till the "ok" node turns
\* "ready" see log-concurrent-leave-last.txt in vE...
\*-------------- This part is added just for simply the proof of HalfNeighbor
\* since we use this action just to prevent the complete down case,
\* we don't really care how it match the original protocol, because it was not 
\* described.
/\ ~\E ms \in receivedMsgs: ms.mreq.type = "ProbeReply" 
/\ ~\E mk \in receivedMsgs: mk.mreq.type = "Probe" /\ mk.destination # i
/\ \A mj \in receivedMsgs: mj.mreq.type = "JoinReply"
                           => GetLSetContent(mj.mreq.lset) = {i}
/\ ~\E mb \in receivedMsgs: mb.mreq.type = "LeaseReply"
\*----------
/\ status' = [status EXCEPT ![i] = "ready"]
/\ lset'= [lset EXCEPT ![i] = EmptyLS(i)]
/\ probing' = [probing EXCEPT ![i] =  {}]
/\ lease' = [lease EXCEPT ![i] = {}]
/\ grant' = [grant EXCEPT ![i] = {}]
/\ failed' = [failed EXCEPT ![i] = {}]
/\ UNCHANGED <<receivedMsgs, toj>>               
     *)
(*     
LSRepair(i) == 
\* find the new candidate for the missing entry in leaf set and try to complete it
/\ status[i] = "wait"
/\ probing[i] = {}
/\ ~IsComplete(lset[i])
/\ Lenth(lset[i].left) # 0  /\ Lenth(lset[i].right) # 0 
/\ LET
   lm == {LeftMost(lset[i])}
   rm == {RightMost(lset[i])}
   newprob == IF Lenth(lset[i].left) < L /\ Lenth(lset[i].right) < L
              THEN lm \cup rm
              ELSE IF Lenth(lset[i].left) < L /\ Lenth(lset[i].right) = L
                   THEN lm
                   ELSE rm \* here it contains the case when leaf set are complete 
                           \* because of overlap happens
   IN 
   /\ probing' = [probing EXCEPT ![i] =  probing[i] \cup newprob]
   /\ receivedMsgs' = receivedMsgs \cup ProbeSet(i, lset[i],  newprob)
/\ UNCHANGED <<lset, failed, lease, status, grant, toj>>
*)
----------------------------------------------------------------
\* (5) Failure simulation actions
MsgLost == 
/\\E m \in receivedMsgs:
   \*/\ PrintT(<<"MsgLost", m>>)
   /\ receivedMsgs' = receivedMsgs \ {m}
/\ UNCHANGED <<status, lset, probing, failed, lease, grant, toj>>
NodeLeft(i) ==
  /\ status[i] = "ready" \* for the moment we assume only ready node can leave
  \*/\ status[i] # "dead" \* a dead node has already left
  \*/\ \E j \in I: j # i /\ status[j] = "ready"
  /\ Cardinality(ReadyNodes) > L + 1\*to make sure there is at lease L+1 nodes working
  /\ status' = [status EXCEPT ![i] = "dead"]
  /\ lset' = [lset EXCEPT ![i] = EmptyLS(i)]
  /\ probing' = [probing EXCEPT ![i] = {}]
  /\ failed' = [failed EXCEPT ![i] = {}]
  /\ lease' = [lease EXCEPT ![i]= {}]
  /\ grant'= [grant EXCEPT ![i]= {}]
  /\ toj' = [toj EXCEPT ![i] = i]
  /\ receivedMsgs' = receivedMsgs \ {m \in receivedMsgs: m.destination = i}
\*  /\ UNCHANGED <<receivedMsgs>>
\* If we delete all the messages destinating i, it means, the messages for i will be ignored, 
\* If we do not delete them, it means, they are still in the pool, even if the node dropped off.
\* The second case is more realistic, because there could be delayed message coming to some node
\* which has dropped off but came again. But of course, there could be lower level mechanism like
\* freshness id to avoid such situation.
==============================================================